from __future__ import annotations

import json
import logging
from abc import abstractmethod

from flask import Response, jsonify
from flask.views import MethodView

import ckan.plugins.toolkit as tk

from ckanext.mailcraft_dashboard.table import TableDefinition, GlobalActionHandler

log = logging.getLogger(__name__)


class ApTableView(MethodView):
    def __init__(
        self,
        table: type[TableDefinition],
        breadcrumb_label: str = "Table",
        page_title: str = "Table",
    ):
        """A generic view to render tables

        Args:
            table: a table definition
            render_template (optional): a path to a render template
        """
        self.table = table
        self.breadcrumb_label = breadcrumb_label
        self.page_title = page_title

    def get(self) -> str | Response:
        """Render a table

        If the data argument is provided, returns the table data
        """
        table = self.table()  # type: ignore

        if tk.request.args.get("data"):
            return jsonify(table.get_data())

        return table.render_table(
            breadcrumb_label=self.breadcrumb_label, page_title=self.page_title
        )

    def post(self) -> Response:
        """Handle global actions on a table"""
        global_action = tk.request.form.get("global_action")
        rows = tk.request.form.get("rows")

        action_func = self.get_global_action(global_action) if global_action else None

        if not action_func or not rows:
            return jsonify(
                {
                    "success": False,
                    "errors": [tk._("The global action is not implemented")],
                }
            )

        errors = []

        for row in json.loads(rows):
            success, error = action_func(row)

            if not success:
                log.debug("Error during global action %s: %s", global_action, error)
                errors.append(error)

        return jsonify({"success": not errors, "errors": errors})

    @abstractmethod
    def get_global_action(self, value: str) -> GlobalActionHandler | None:
        pass



from __future__ import annotations

import logging
from typing import Any

from sqlalchemy import Column, DateTime, Integer, Text, func
from sqlalchemy.orm import Query
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.mutable import MutableDict
from typing_extensions import Self

import ckan.model as model
from ckan.plugins import toolkit as tk

from ckanext.mailcraft.types import EmailData

log = logging.getLogger(__name__)


class Email(tk.BaseModel):
    __tablename__ = "mailcraft_mail"

    class State:
        failed = "failed"
        success = "success"
        stopped = "stopped"

    id = Column(Integer, primary_key=True)

    subject = Column(Text)
    timestamp = Column(DateTime, nullable=False, default=func.now())
    sender = Column(Text)
    recipient = Column(Text)
    message = Column(Text)
    state = Column(Text, nullable=False, default=State.success)
    extras = Column("extras", MutableDict.as_mutable(JSONB))

    @classmethod
    def all(cls) -> list[dict[str, Any]]:
        query: Query = model.Session.query.query(cls).order_by(cls.timestamp.desc())

        return [mail.dictize({}) for mail in query.all()]

    @classmethod
    def create(cls, **kwargs) -> Email:
        mail = cls(**kwargs)

        model.Session.add(mail)
        model.Session.commit()

        return mail

    @classmethod
    def save_mail(
        cls,
        email_data: EmailData,
        body_html: str,
        state: str,
    ) -> Email:
        mail = cls(
            subject=email_data["Subject"],
            timestamp=email_data["Date"],
            sender=email_data["From"],
            recipient=email_data["To"],
            message=body_html,
            state=state,
            extras=email_data,
        )

        model.Session.add(mail)
        model.Session.commit()

        return mail

    def dictize(self, context):
        return {
            "id": self.id,
            "subject": self.subject,
            "timestamp": self.timestamp.isoformat(),
            "sender": self.sender,
            "recipient": self.recipient,
            "message": self.message,
            "state": self.state,
            "extras": self.extras or {},
        }

    @classmethod
    def clear_emails(cls) -> int:
        rows_deleted = model.Session.query(cls).delete()
        model.Session.commit()

        return rows_deleted

    @classmethod
    def get(cls, mail_id: str) -> Self | None:
        query: Query = model.Session.query(cls).filter(cls.id == mail_id)

        return query.one_or_none()

    def delete(self) -> None:
        model.Session().autoflush = False
        model.Session.delete(self)



from __future__ import annotations

import ckan.plugins as plugins
import ckan.plugins.toolkit as toolkit
from ckan.common import CKANConfig

import ckanext.mailcraft.config as mc_config
from ckanext.mailcraft.mailer import DefaultMailer


@toolkit.blanket.config_declarations
class MailcraftPlugin(plugins.SingletonPlugin):
    plugins.implements(plugins.IConfigurer)
    plugins.implements(plugins.IConfigurable)

    # IConfigurer

    def update_config(self, config_):
        toolkit.add_template_directory(config_, "templates")
        toolkit.add_public_directory(config_, "public")

    # IConfigurable

    def configure(self, config: CKANConfig) -> None:
        if mc_config.is_startup_conn_test_enabled():
            DefaultMailer().test_conn()



from typing import Type

from ckan.plugins.interfaces import Interface

from ckanext.mailcraft.mailer import DefaultMailer


class IMailCraft(Interface):
    """Allow to register a custom mailer instead of a default one"""

    def get_mailer(self) -> Type[DefaultMailer]:
        """Return the mailer class"""
        return DefaultMailer



import ckan.plugins as p

from ckanext.mailcraft.mailer import DefaultMailer
from ckanext.mailcraft.interfaces import IMailCraft


def get_mailer() -> DefaultMailer:
    mailer = DefaultMailer

    for plugin in reversed(list(p.PluginImplementations(IMailCraft))):
        mailer = plugin.get_mailer()

    return mailer()



from __future__ import annotations

from typing import IO

from typing_extensions import NotRequired, TypedDict

AttachmentWithType = tuple[str, IO[str], str] | tuple[str, IO[bytes], str]
AttachmentWithoutType = tuple[str, IO[str]] | tuple[str, IO[bytes]]
Attachment = AttachmentWithType | AttachmentWithoutType


EmailData = TypedDict(
    "EmailData",
    {
        "Bcc": str,
        "Content-Type": NotRequired[str],
        "Date": str,
        "From": str,
        "MIME-Version": NotRequired[str],
        "Subject": str,
        "To": str,
        "X-Mailer": NotRequired[str],
        "redirected_from": NotRequired["list[str]"],
    },
)



from typing import Any, Callable

import pytest

import ckan.plugins.toolkit as tk
from ckan.tests.helpers import call_action


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestCreateMail:
    def test_create_mail(self, email_factory: Callable[..., dict[str, Any]]):
        mail = email_factory()

        assert mail is not None
        assert mail["subject"] is not None
        assert mail["sender"] is not None
        assert mail["recipient"] is not None
        assert mail["message"] is not None
        assert mail["state"] is not None
        assert mail["extras"] is not None


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestListMails:
    def test_list_mails(self, email: dict[str, Any]):
        assert call_action("mc_mail_list")

    def test_no_mails(self):
        assert call_action("mc_mail_list") == []


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestShowMail:
    def test_show_mail(self, email: dict[str, Any]):
        assert call_action("mc_mail_show", {}, id=email["id"]) == email

    def test_no_mail(self):
        with pytest.raises(tk.ValidationError):
            call_action("mc_mail_show", id="123")


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestDeleteMail:
    def test_delete_mail(self, email: dict[str, Any]):
        assert call_action("mc_mail_delete", {}, id=email["id"])

    def test_no_mail(self):
        with pytest.raises(tk.ValidationError):
            call_action("mc_mail_delete", {}, id="123")


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestClearMails:
    def test_clear_no_mails(self):
        assert call_action("mc_mail_clear")["deleted"] == 0

    def test_clear_mails(self, email_factory: Callable[..., dict[str, Any]]):
        for _ in range(10):
            email_factory()

        assert len(call_action("mc_mail_list")) == 10
        assert call_action("mc_mail_clear")["deleted"] == 10
        assert call_action("mc_mail_list") == []



from __future__ import annotations

import codecs
import logging
import mimetypes
import os
import smtplib
from abc import ABC, abstractmethod
from collections.abc import Iterable
from email import utils as email_utils
from email.message import EmailMessage
from time import time
from typing import Any, Optional, cast

import ckan.plugins as p
import ckan.plugins.toolkit as tk
from ckan import model

import ckanext.mailcraft.config as mc_config
import ckanext.mailcraft_dashboard.model as mc_model
from ckanext.mailcraft.exception import MailerException
from ckanext.mailcraft.types import (
    Attachment,
    AttachmentWithoutType,
    AttachmentWithType,
    EmailData,
)

log = logging.getLogger(__name__)

ATTACHMENT_WITH_TYPE_LEN = 3


class BaseMailer(ABC):
    """Base class for mailers."""

    def __init__(self):
        """Initialize the mailer with SMTP settings from CKAN config."""
        self.server = tk.config["smtp.server"]
        self.start_tls = tk.asbool(tk.config["smtp.starttls"])
        self.user = tk.config["smtp.user"]
        self.password = tk.config["smtp.password"]
        self.mail_from = tk.config["smtp.mail_from"]
        self.reply_to = tk.config["smtp.reply_to"]

        self.site_title = tk.config["ckan.site_title"]
        self.site_url = tk.config["ckan.site_url"]

        self.conn_timeout = mc_config.get_conn_timeout()
        self.stop_outgoing = mc_config.stop_outgoing_emails()
        self.save_emails = mc_config.save_emails_to_db()
        self.redirect_to = mc_config.get_redirect_email()

    @abstractmethod
    def mail_recipients(  # noqa: PLR0913
        self,
        subject: str,
        recipients: list[str],
        body: str,
        body_html: str,
        headers: dict[str, Any] | None = None,
        attachments: Iterable[Attachment] | None = None,
        to: list[str] | None = None,
    ):
        """Send an email to a list of recipients."""

    @abstractmethod
    def add_attachments(
        self, msg: EmailMessage, attachments: Iterable[Attachment]
    ) -> None:
        """Add attachments to the email message."""

    @abstractmethod
    def get_connection(self) -> smtplib.SMTP:
        """Get an SMTP connection object."""

    @abstractmethod
    def test_conn(self):
        """Test the SMTP connection."""

    @abstractmethod
    def mail_user(  # noqa: PLR0913
        self,
        user: str,
        subject: str,
        body: str,
        body_html: str,
        headers: dict[str, Any] | None = None,
        attachments: Iterable[Attachment] | None = None,
    ) -> None:
        """Send an email to a CKAN user by their ID or name."""


class DefaultMailer(BaseMailer):
    """Default mailer implementation."""
    def mail_recipients(  # noqa: PLR0913, C901
        self,
        subject: str,
        recipients: list[str],
        body: str,
        body_html: str,
        headers: dict[str, Any] | None = None,
        attachments: Iterable[Attachment] | None = None,
        to: list[str] | None = None,
    ):
        """Send an email to a list of recipients."""
        headers = headers or {}
        attachments = attachments or []

        if self.redirect_to:
            log.info("Redirecting email to %s instead of %s", self.redirect_to, recipients)
            recipients = self.redirect_to

        msg = EmailMessage()

        msg["From"] = email_utils.formataddr((self.site_title, self.mail_from))
        msg["Subject"] = subject
        msg["Date"] = email_utils.formatdate(time())
        msg["To"] = ", ".join(to) if to else  ", ".join(recipients)


        if not tk.config.get("ckan.hide_version"):
            msg["X-Mailer"] = f"CKAN {tk.h.ckan_version()}"

        for k, v in headers.items():
            msg.replace_header(k, v) if k in msg else msg.add_header(k, v)

        # Assign Reply-to if configured and not set via headers
        if self.reply_to and not msg["Reply-to"]:
            msg["Reply-to"] = self.reply_to

        msg.set_content(body, cte="base64")
        msg.add_alternative(body_html, subtype="html", cte="base64")

        if attachments:
            self.add_attachments(msg, attachments)

        email_data: EmailData = dict(msg.items())  # type: ignore

        try:
            if self.stop_outgoing:
                self._save_email(email_data, body_html, mc_model.Email.State.stopped)
            else:
                self._send_email(recipients, msg)
        except MailerException:
            log.exception("Error sending email to %s", recipients)
            self._save_email(email_data, body_html, mc_model.Email.State.failed)
        else:
            if not self.stop_outgoing:
                self._save_email(email_data, body_html)

    def add_attachments(
        self, msg: EmailMessage, attachments: Iterable[Attachment]
    ) -> None:
        """Add attachments to the email message."""
        for attachment in attachments:
            if len(attachment) == ATTACHMENT_WITH_TYPE_LEN:
                name, _file, media_type = cast(AttachmentWithType, attachment)
            else:
                name, _file = cast(AttachmentWithoutType, attachment)
                media_type = None

            if not media_type:
                media_type, _ = mimetypes.guess_type(name)

            main_type, sub_type = media_type.split("/") if media_type else (None, None)

            msg.add_attachment(
                _file.read(),
                filename=name,
                maintype=main_type,
                subtype=sub_type,
            )

    def get_connection(self) -> smtplib.SMTP:
        """Get an SMTP conn object."""
        try:
            conn = smtplib.SMTP(self.server, timeout=self.conn_timeout)
        except OSError as e:
            log.exception('SMTP server could not be connected to: "%s"', self.server)
            raise MailerException(  # noqa: TRY003
                'SMTP server could not be connected to: "%s" %s', self.server, e
            ) from e

        try:
            conn.ehlo()

            if self.start_tls:
                if conn.has_extn("STARTTLS"):
                    conn.starttls()
                    conn.ehlo()
                else:
                    raise MailerException("SMTP server does not support STARTTLS") # noqa: TRY003

            if self.user:
                conn.login(self.user, self.password)
        except smtplib.SMTPException as e:
            log.exception("An error occurred during SMTP authentication")
            raise MailerException(f"{e}") from e

        return conn

    def _save_email(
        self,
        email_data: EmailData,
        body_html: str,
        state: str = mc_model.Email.State.success,
    ) -> None:
        if not p.plugin_loaded("mailcraft_dashboard") or not self.save_emails:
            return

        mc_model.Email.save_mail(email_data, body_html, state)

    def _send_email(self, recipients: list[str], msg: EmailMessage):
        conn = self.get_connection()

        try:
            conn.sendmail(self.mail_from, recipients, msg.as_string())
            log.info("Sent email to %s", recipients)
        except smtplib.SMTPException as e:
            log.exception("Error sending email: %s")
            raise MailerException("Error sending email: %s", e) from e # noqa: TRY003
        finally:
            conn.quit()

    def test_conn(self):
        """Test the SMTP connection."""
        conn = self.get_connection()
        conn.quit()

    def mail_user(  # noqa: PLR0913
        self,
        user: str,
        subject: str,
        body: str,
        body_html: str,
        headers: dict[str, Any] | None = None,
        attachments: Iterable[Attachment] | None = None,
    ) -> None:
        """Sends an email to a CKAN user by its ID or name."""
        user_obj = model.User.get(user)

        if not user_obj:
            raise MailerException(tk._("User doesn't exist"))

        if not user_obj.email:
            raise MailerException(tk._("User doesn't have an email address"))

        self.mail_recipients(
            subject,
            [user_obj.email],
            body,
            body_html=body_html,
            headers=headers,
            attachments=attachments,
        )

    def send_reset_link(self, user: model.User) -> None:
        """Sends a password reset link to a user.

        Args:
            user: The user to send the reset link to.
        """
        self.create_reset_key(user)
        extra_vars = {
            "reset_link": tk.h.url_for(
                "user.perform_reset", id=user.id, key=user.reset_key, qualified=True
            ),
            "site_title": tk.config.get("ckan.site_title"),
            "site_url": tk.config.get("ckan.site_url"),
            "user_name": user.name,
        }

        self.mail_user(
            user=user.name,
            subject="Reset your password",
            body=tk.render(
                "mailcraft/emails/reset_password/body.txt",
                extra_vars,
            ),
            body_html=tk.render(
                "mailcraft/emails/reset_password/body.html",
                extra_vars,
            ),
        )

    def create_reset_key(self, user: model.User):
        """Creates a reset key for a user and saves it to the database."""
        user.reset_key = codecs.encode(os.urandom(16), "hex").decode()
        model.repo.commit_and_remove()

    def verify_reset_link(self, user: model.User, key: str | None) -> bool:
        """Verifies if the reset key is valid for the user."""
        if not key or not user.reset_key:
            return False

        return key.strip() == user.reset_key



from ckan import types


def mc_mail_create(context: types.Context, data_dict: types.DataDict):
    return {"success": False}


def mc_mail_list(context: types.Context, data_dict: types.DataDict):
    return {"success": False}


def mc_mail_show(context: types.Context, data_dict: types.DataDict):
    return {"success": False}


def mc_mail_delete(context: types.Context, data_dict: types.DataDict):
    return {"success": False}



from __future__ import annotations

import ckan.plugins as plugins
import ckan.plugins.toolkit as toolkit


@toolkit.blanket.blueprints
@toolkit.blanket.actions
@toolkit.blanket.auth_functions
@toolkit.blanket.validators
@toolkit.blanket.helpers
class MailcraftDashboardPlugin(plugins.SingletonPlugin):
    plugins.implements(plugins.IConfigurer)

    # IConfigurer

    def update_config(self, config_):
        toolkit.add_resource("assets", "mailcraft_dashboard")
        toolkit.add_template_directory(config_, "templates")



from __future__ import annotations

import uuid
from abc import abstractmethod
from collections.abc import Callable
from datetime import datetime
from typing import Any, TypeAlias

import ckan.plugins.toolkit as tk

Value: TypeAlias = Any
Options: TypeAlias = "dict[str, Any]"
Row: TypeAlias = dict[str, Any]
GlobalActionHandlerResult: TypeAlias = tuple[bool, str | None]
GlobalActionHandler: TypeAlias = Callable[[Row], GlobalActionHandlerResult]
FormatterResult: TypeAlias = str


def date(
    value: datetime,
    options: dict[str, Any],
    name: str,
    record: Any,
    table: TableDefinition,
) -> str:
    """Render a datetime object as a string.

    Args:
        value (datetime): date value
        options: options for the renderer
        name (str): column name
        record (Any): row data
        table: table definition

    Options:
        - `date_format` (str) - date format string. **Default** is `%d/%m/%Y - %H:%M`

    Returns:
        formatted date
    """
    date_format: str = options.get("date_format", "%d/%m/%Y - %H:%M")

    return tk.h.render_datetime(value, date_format=date_format)


def actions(
    value: Value,
    options: Options,
    column: ColumnDefinition,
    row: Row,
    table: TableDefinition,
) -> FormatterResult:
    """Render actions for the table row.

    Args:
        value: string value
        options: options for the renderer
        column: column definition
        row: row data
        table: table definition

    Options:
        - `template` (str) - template to render the actions.
    """
    template = options.get("template", "mailcraft/tables/formatters/actions.html")

    return tk.literal(
        tk.render(
            template,
            extra_vars={"table": table, "column": column, "row": row},
        )
    )


class TableDefinition:
    """Defines a table to be rendered with Tabulator."""

    def __init__(  # noqa: PLR0913
        self,
        name: str,
        ajax_url: str,
        columns: list[ColumnDefinition] | None = None,
        actions: list[ActionDefinition] | None = None,
        global_actions: list[GlobalActionDefinition] | None = None,
        placeholder: str | None = None,
        pagination: bool = True,
        page_size: int = 10,
        selectable: bool = False,
        table_action_snippet: str | None = None,
        table_template: str = "mailcraft/tables/table_base.html",
    ):
        """Initialize a table definition.

        Args:
            name (str): Unique identifier for the table
            ajax_url (str): URL to fetch data from
            columns (list, optional): List of ColumnDefinition objects
            actions (list, optional): List of ActionDefinition objects
            global_actions (list, optional): List of GlobalActionDefinition objects
            placeholder (str, optional): Placeholder text for the table
            pagination (bool): Whether to enable pagination
            page_size (int): Number of rows per page
            selectable (bool): Whether rows can be selected
            table_action_snippet (str, optional): Snippet to render table actions
            table_template (str, optional): Template to render the table
        """
        self.id = f"table_{name}_{uuid.uuid4().hex[:8]}"
        self.name = name
        self.ajax_url = ajax_url
        self.columns = columns or []
        self.actions = actions or []
        self.global_actions = global_actions or []
        self.placeholder = placeholder or "No emails found"
        self.pagination = pagination
        self.page_size = page_size
        self.selectable = True if self.global_actions else selectable
        self.table_action_snippet = table_action_snippet
        self.table_template = table_template

    def get_tabulator_config(self) -> dict[str, Any]:
        """Return the Tabulator configuration for the table."""
        columns = [col.to_dict() for col in self.columns]

        options = {
            "columns": columns,
            "layout": "fitDataFill",
            "placeholder": self.placeholder,
            "ajaxURL": self.ajax_url,
        }

        if self.pagination:
            options["pagination"] = "local"
            options["paginationSize"] = self.page_size
            options["paginationSizeSelector"] = [5, 10, 25, 50, 100]

        if self.selectable or self.global_actions:
            options["selectableRows"] = True
            options["selectableRangeMode"] = "click"
            options["selectableRollingSelection"] = False
            options["selectablePersistence"] = False

        return options

    def render_table(self, **kwargs: Any) -> str:
        """Render the table template with the necessary data."""
        return tk.render(self.table_template, extra_vars={"table": self, **kwargs})

    @abstractmethod
    def get_raw_data(self) -> list[dict[str, Any]]:
        """Return the list of rows to be rendered in the table.

        Returns:
            list[dict[str, Any]]: List of rows to be rendered in the table
        """

    def get_data(self) -> list[Any]:
        """Get the data for the table with applied formatters."""
        self._formatters = self.get_formatters()

        return [self.apply_formatters(dict(row)) for row in self.get_raw_data()]

    def get_formatters(self) -> dict[str, Callable[..., Any]]:
        """Return a dict of available formatters."""
        return {"date": date, "actions": actions}

    def apply_formatters(self, row: dict[str, Any]) -> dict[str, Any]:
        """Apply formatters to each cell in a row."""
        for column in self.columns:
            cell_value = row.get(column.field)

            if not column.formatters:
                continue

            for formatter, formatter_options in column.formatters:
                formatter_function = self._formatters[formatter]

                cell_value = formatter_function(
                    cell_value, formatter_options, column, row, self
                )

            row[column.field] = cell_value

        return row


class ColumnDefinition:
    """Defines how a column should be rendered in Tabulator."""

    def __init__(  # noqa: PLR0913
        self,
        field: str,
        title: str | None = None,
        formatters: list[tuple[str, dict[str, Any]]] | None = None,
        tabulator_formatter: str | None = None,
        tabulator_formatter_params: dict[str, Any] | None = None,
        width: int | None = None,
        min_width: int | None = None,
        visible: bool = True,
        sorter: str | None = "string",
        filterable: bool = True,
        resizable: bool = True,
    ):
        """Initialize a column definition.

        Args:
            field (str): The field name in the data dict
            title (str, optional): The display title for the column
            formatters (list, optional): List of formatters to apply to the column
            tabulator_formatter (str, optional): Tabulator formatter to apply to the column
            tabulator_formatter_params (dict, optional): Parameters for the tabulator formatter
            width (int, optional): Width of the column
            min_width (int, optional): Minimum width of the column
            visible (bool): Whether the column is visible
            sorter (str, optional): Default sorter for the column
            filterable (bool): Whether the column can be filtered
            resizable (bool): Whether the column is resizable
        """
        self.field = field
        self.title = title or field.replace("_", " ").title()
        self.formatters = formatters
        self.tabulator_formatter = tabulator_formatter
        self.tabulator_formatter_params = tabulator_formatter_params
        self.width = width
        self.min_width = min_width
        self.visible = visible
        self.sorter = sorter
        self.filterable = filterable
        self.resizable = resizable

    def __repr__(self):
        """String representation of the column definition."""
        return f"ColumnDefinition(field={self.field}, title={self.title})"

    def to_dict(self) -> dict[str, Any]:
        """Convert the column definition to a dict for JSON serialization."""
        result = {
            "field": self.field,
            "title": self.title,
            "visible": self.visible,
            "resizable": self.resizable,
        }

        if self.sorter:
            result["sorter"] = self.sorter
        else:
            result["headerSort"] = False

        if self.tabulator_formatter:
            result["formatter"] = self.tabulator_formatter

        if self.tabulator_formatter_params:
            result["formatterParams"] = self.tabulator_formatter_params

        if self.width:
            result["width"] = self.width

        if self.min_width:
            result["minWidth"] = self.min_width

        return result


class ActionDefinition:
    """Defines an action that can be performed on a row."""

    def __init__(  # noqa: PLR0913
        self,
        name: str,
        label: str | None = None,
        icon: str | None = None,
        url: str | None = None,
        endpoint: str | None = None,
        url_params: dict[str, Any] | None = None,
        css_class: str | None = None,
        visible_callback: Callable[..., bool] | None = None,
        attrs: dict[str, Any] | None = None,
    ):
        """Initialize an action definition.

        Args:
            name (str): Unique identifier for the action
            label (str, optional): Display label for the action
            icon (str, optional): Icon class (e.g., "fa fa-edit")
            url (str, optional): Static URL for the action
            endpoint (str, optional): Flask endpoint to generate URL
            url_params (dict, optional): Parameters for the URL
            css_class (str, optional): CSS class for styling
            visible_callback (callable, optional): Function that determines if action is visible
            attrs (dict, optional): Additional attributes for the action
        """
        self.name = name
        self.label = label
        self.icon = icon
        self.url = url
        self.endpoint = endpoint
        self.url_params = url_params
        self.css_class = css_class
        self.visible_callback = visible_callback
        self.attrs = attrs or {}

    def __repr__(self):
        """String representation of the action definition."""
        return f"ActionDefinition(name={self.name})"

    def to_dict(self, row_data: Any | None = None):
        """Convert the action definition to a dict for JSON serialization."""
        # Check if action should be visible for this row
        if self.visible_callback and row_data and not self.visible_callback(row_data):
            return None

        result = {
            "name": self.name,
            "label": self.label,
            "attrs": self.attrs,
        }

        if self.icon:
            result["icon"] = self.icon

        if self.css_class:
            result["cssClass"] = self.css_class

        return result


class GlobalActionDefinition:
    """Defines an action that can be performed on multiple rows."""

    def __init__(
        self,
        action: str,
        label: str,
    ):
        """Initialize a global action definition.

        Args:
            action (str): Unique identifier for the action
            label (str): Display label for the action
        """
        self.action = action
        self.label = label

    def __repr__(self):
        """String representation of the global action definition."""
        return f"GlobalActionDefinition(action={self.action}, label={self.label})"

    def to_dict(self):
        """Convert the global action definition to a dict."""
        return {
            "action": self.action,
            "label": self.label,
        }



from __future__ import annotations

from typing import Any

from ckanext.mailcraft.utils import get_mailer
from ckanext.mailcraft_dashboard.generics import ApTableView
from ckanext.mailcraft_dashboard.table import (
    ActionDefinition,
    ColumnDefinition,
    GlobalActionDefinition,
    GlobalActionHandler,
    GlobalActionHandlerResult,
    Row,
    TableDefinition,
)

import ckan.plugins.toolkit as tk
from ckan import types
from flask import Blueprint, Response
from flask.views import MethodView

mailcraft = Blueprint("mailcraft", __name__, url_prefix="/ckan-admin/mailcraft")


def before_request() -> None:
    """A before request handler to check for sysadmin rights."""
    try:
        tk.check_access("sysadmin", {"user": tk.current_user.name})
    except tk.NotAuthorized:
        tk.abort(403, tk._("Need to be system administrator to administer"))


class DashboardTable(TableDefinition):
    """Table definition for the mailcraft dashboard."""
    def __init__(self):
        """Initialize the table definition."""
        super().__init__(
            name="content",
            ajax_url=tk.url_for("mailcraft.dashboard", data=True),
            columns=[
                ColumnDefinition(field="id", filterable=False, resizable=False),
                ColumnDefinition(field="subject", width=250),
                ColumnDefinition(field="sender"),
                ColumnDefinition(field="recipient"),
                ColumnDefinition(field="state", resizable=False, width=100),
                ColumnDefinition(
                    field="timestamp",
                    formatters=[("date", {"date_format": "%Y-%m-%d %H:%M"})],
                    resizable=False,
                ),
                ColumnDefinition(
                    field="actions",
                    formatters=[("actions", {})],
                    filterable=False,
                    tabulator_formatter="html",
                    sorter=None,
                    resizable=False,
                ),
            ],
            actions=[
                ActionDefinition(
                    name="view",
                    label=tk._("View"),
                    icon="fa fa-eye",
                    endpoint="mailcraft.mail_read",
                    url_params={
                        "view": "read",
                        "mail_id": "$id",
                    },
                ),
            ],
            global_actions=[
                GlobalActionDefinition(
                    action="delete", label="Delete selected entities"
                ),
            ],
            table_action_snippet="mailcraft/table_actions.html",
        )

    def get_raw_data(self) -> list[dict[str, Any]]:
        """Fetch raw data for the table."""
        return tk.get_action("mc_mail_list")(_build_context(), {})


class DashboardView(ApTableView):
    """View for the mailcraft dashboard."""
    def get_global_action(self, value: str) -> GlobalActionHandler | None:
        """Return the handler for a global action."""
        return {"delete": self._remove_emails}.get(value)

    @staticmethod
    def _remove_emails(row: Row) -> GlobalActionHandlerResult:
        try:
            tk.get_action("mc_mail_delete")(
                {"ignore_auth": True},
                {"id": row["id"]},
            )
        except tk.ObjectNotFound:
            return False, tk._("Mail not found")

        return True, None


class MailReadView(MethodView):
    """View for reading a single email."""
    def get(self, mail_id: str) -> str:
        """Render the email reading template."""
        try:
            mail = tk.get_action("mc_mail_show")(_build_context(), {"id": mail_id})
        except tk.ValidationError:
            return tk.render("mailcraft/404.html")

        return tk.render("mailcraft/mail_read.html", extra_vars={"mail": mail})


class MailClearView(MethodView):
    """View for clearing all emails."""
    def post(self) -> Response:
        """Clear all emails and redirect to the dashboard."""
        tk.get_action("mc_mail_clear")(_build_context(), {})

        return tk.redirect_to("mailcraft.dashboard")


class MailTestView(MethodView):
    """View for sending a test email."""
    def post(self) -> Response:
        """Send a test email and redirect to the dashboard."""
        mailer = get_mailer()

        mailer.mail_recipients(
            subject="Hello world",
            recipients=["test@gmail.com"],
            body="Hello world",
            body_html=tk.render(
                "mailcraft/emails/test.html",
                extra_vars={
                    "site_url": mailer.site_url,
                    "site_title": mailer.site_title,
                },
            ),
        )

        return tk.redirect_to("mailcraft.dashboard")


def _build_context() -> types.Context:
    return {
        "user": tk.current_user.name,
        "auth_user_obj": tk.current_user,
    }


mailcraft.add_url_rule("/test", view_func=MailTestView.as_view("send_test"))
mailcraft.add_url_rule(
    "/dashboard",
    view_func=DashboardView.as_view(
        "dashboard",
        table=DashboardTable,
        page_title="",
    ),
)
mailcraft.add_url_rule(
    "/dashboard/read/<mail_id>", view_func=MailReadView.as_view("mail_read")
)
mailcraft.add_url_rule(
    "/dashboard/clear", view_func=MailClearView.as_view("clear_mails")
)



import json
from typing import Any

import ckan.plugins.toolkit as tk


def mc_dumps(value: Any) -> str:
    """Convert a value to a JSON string.

    Args:
        value: The value to convert to a JSON string

    Returns:
        The JSON string
    """
    return json.dumps(value)


def mc_build_url_from_params(
    endpoint: str, url_params: dict[str, Any], row: dict[str, Any]
) -> str:
    """Build an action URL based on the endpoint and URL parameters.

    The url_params might contain values like $id, $type, etc.
    We need to replace them with the actual values from the row

    Args:
        endpoint: The endpoint to build the URL for
        url_params: The URL parameters to build the URL for
        row: The row to build the URL for
    """
    params = url_params.copy()

    for key, value in params.items():
        if value.startswith("$"):
            params[key] = row[value[1:]]

    return tk.url_for(endpoint, **params)
